/***************************************************************************
 * Copyright 2013 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.service.impl;

import java.util.ArrayList;
import java.util.List;

import org.springframework.stereotype.Service;

import kieker.webgui.common.exception.GraphLayoutException;
import kieker.webgui.service.IGraphLayoutService;

import org.eclipse.emf.common.util.EList;

import de.cau.cs.kieler.core.alg.BasicProgressMonitor;
import de.cau.cs.kieler.core.alg.IKielerProgressMonitor;
import de.cau.cs.kieler.core.kgraph.KEdge;
import de.cau.cs.kieler.core.kgraph.KLabeledGraphElement;
import de.cau.cs.kieler.core.kgraph.KNode;
import de.cau.cs.kieler.core.kgraph.KPort;
import de.cau.cs.kieler.kiml.AbstractLayoutProvider;
import de.cau.cs.kieler.kiml.klayoutdata.KEdgeLayout;
import de.cau.cs.kieler.kiml.klayoutdata.KPoint;
import de.cau.cs.kieler.kiml.klayoutdata.KShapeLayout;
import de.cau.cs.kieler.kiml.options.EdgeRouting;
import de.cau.cs.kieler.kiml.options.LayoutOptions;
import de.cau.cs.kieler.kiml.options.PortConstraints;
import de.cau.cs.kieler.kiml.util.KimlUtil;
import de.cau.cs.kieler.klay.layered.LayeredLayoutProvider;

/**
 * This is a utility class providing a method to layout a graph from the graph flow editor. The method accepts two strings (one for the nodes, one for the edges)
 * which are the result of specific javascript functions of the graph. For further information look into the API description of the flow editor itself.
 * 
 * @author row, Nils Christian Ehmke
 */
@Service
public final class GraphLayoutServiceImpl implements IGraphLayoutService {

	/**
	 * Default constructor. <b>Do not use this constructor. This bean is Spring managed.</b>
	 */
	public GraphLayoutServiceImpl() {
		// No code necessary
	}

	private static void addNodes(final int scale, final String[] nodeInfo, final LayoutInformation layoutInformation) throws NumberFormatException {
		// define node and port list
		final List<List<KLabeledGraphElement>> nodesAndPorts = layoutInformation.getNodesAndPorts();
		List<KLabeledGraphElement> nodeAndPortList;

		// retrieve size modifier
		final int dimHalf = scale;
		final int dim = dimHalf * 2;

		// loop variables //
		String[] nodeProps;
		KNode node;
		int parentIndex;
		KNode parent;
		KPort port;
		KShapeLayout layout;

		boolean isNodeFamily;
		int width;
		int height;
		int portCount;
		int portType;
		int portX;
		int portY;

		int i;
		int n;
		int p;
		int lp;
		final int l = nodeInfo.length;

		// -- 1. initialize empty nodes --
		for (n = 0; n < l; n++) {
			// init node
			node = KimlUtil.createInitializedNode();

			// create a list that will contain the node itself
			// as well as its ports (should they exist)
			nodeAndPortList = new ArrayList<KLabeledGraphElement>();
			nodeAndPortList.add(node);

			// add the list to the global 2 dimensional list
			nodesAndPorts.add(n, nodeAndPortList);
		}

		// -- 2. Define nodes by the properties, given in the string --
		for (n = 0; n < l; n++) {
			nodeProps = nodeInfo[n].split(" ");

			// get Node from global list
			node = (KNode) nodesAndPorts.get(n).get(0);

			// i is our property array-pointer
			i = 0;

			// retrieve node information
			isNodeFamily = "f".equals(nodeProps[i++]);
			width = Integer.parseInt(nodeProps[i++]);
			height = Integer.parseInt(nodeProps[i++]);

			// determine the parent of the node
			parentIndex = Integer.parseInt(nodeProps[i++]);
			if (parentIndex > 0) {
				parent = (KNode) nodesAndPorts.get(parentIndex).get(0);
				node.setParent(parent);
			} else {
				node.setParent(layoutInformation.getGraph());
			}

			// set node dimensions
			layout = node.getData(KShapeLayout.class);
			layout.setWidth(width);
			layout.setHeight(height);

			// parse port data if it is a NodeFamily
			if (isNodeFamily) {

				// let the node size be fixed for now
				// (resizable nodeFamilies may be implemented in the future)
				layout.setProperty(LayoutOptions.PORT_CONSTRAINTS, PortConstraints.FIXED_POS);
				portX = -dimHalf;

				// add Ports
				for (portType = 0; portType < 3; portType++) {
					// determine port x-position
					if (portType == 1) {
						portX = width - dimHalf;
					}

					portCount = Integer.parseInt(nodeProps[i++]);
					portY = Integer.parseInt(nodeProps[i++]);

					for (p = 0, lp = portCount; p < lp; p++) {

						// create port
						port = KimlUtil.createInitializedPort();
						port.setNode(node);
						nodesAndPorts.get(n).add(port);

						// set layout for Port
						layout = port.getData(KShapeLayout.class);

						// set position and dimension of port
						layout.setWidth(dim);
						layout.setHeight(dimHalf);
						layout.setYpos(portY);
						layout.setXpos(portX);

						portY += dim;
					}
				}
			}
		}
	}

	private static void addEdges(final String[] edgeInfo, final LayoutInformation layoutInformation) throws NumberFormatException {
		final List<List<KLabeledGraphElement>> nodesAndPorts = layoutInformation.getNodesAndPorts();

		// set up loop variables
		String[] edgeProps;
		KEdge edge;

		int id;
		boolean hasSourcePort;

		KNode sourceNode;
		int sourcePortIndex;
		KPort sourcePort = null;
		List<KLabeledGraphElement> sourceList;

		int targetPortIndex;
		KPort targetPort;
		List<KLabeledGraphElement> targetList;

		int ea = 0;
		int p;
		int epl;

		// read from String
		for (final String element : edgeInfo) {

			// get edge info of a single source
			edgeProps = element.split(" ");
			p = 0;

			// get source index and port
			id = Integer.parseInt(edgeProps[p++]);
			sourceList = nodesAndPorts.get(id);
			sourceNode = (KNode) sourceList.get(0);

			sourcePortIndex = Integer.parseInt(edgeProps[p++]) + 1;
			hasSourcePort = sourcePortIndex != 0;

			if (hasSourcePort) {
				sourcePort = (KPort) sourceList.get(sourcePortIndex);
			}

			// get target ports
			epl = edgeProps.length;
			while (p < epl) {

				// get target index and port
				id = Integer.parseInt(edgeProps[p++]);
				targetList = nodesAndPorts.get(id);

				// define edge
				edge = KimlUtil.createInitializedEdge();
				edge.setSource(sourceNode);
				edge.setTarget((KNode) targetList.get(0));

				targetPortIndex = Integer.parseInt(edgeProps[p++]) + 1;
				if (targetPortIndex != 0) {
					targetPort = (KPort) targetList.get(targetPortIndex);

					// set target port of edge
					edge.setTargetPort(targetPort);
					targetPort.getEdges().add(edge);
				}
				if (hasSourcePort) {
					// set source port of edge
					edge.setSourcePort(sourcePort);
					sourcePort.getEdges().add(edge);
				}
				layoutInformation.getEdges().add(ea++, edge);
			}

		}
	}

	private static String[] correctEmptyArray(final String[] arr) {
		if ((arr.length == 1) && arr[0].isEmpty()) {
			return new String[0];
		} else {
			return arr;
		}
	}

	private static LayoutInformation assembleLayoutInformation(final String nodesStr, final String edgesStr) throws GraphLayoutException {
		final int firstSemicolon = nodesStr.indexOf(';');

		// if there is no semicolon, we got an empty graph
		if (firstSemicolon < 0) {
			throw new GraphLayoutException("Empty graph");
		}

		final int scale = Integer.parseInt(nodesStr.substring(0, firstSemicolon));

		final String[] nodeInfo = GraphLayoutServiceImpl.correctEmptyArray(nodesStr.substring(firstSemicolon + 1).split(";"));
		final String[] edgeInfo = GraphLayoutServiceImpl.correctEmptyArray(edgesStr.split(";"));

		final List<List<KLabeledGraphElement>> children = new ArrayList<List<KLabeledGraphElement>>(nodeInfo.length - 1);
		final List<KEdge> edges = new ArrayList<KEdge>(edgeInfo.length);
		final LayoutInformation layoutInformation = new LayoutInformation(KimlUtil.createInitializedNode(), children, edges);

		GraphLayoutServiceImpl.addNodes(scale, nodeInfo, layoutInformation);
		GraphLayoutServiceImpl.addEdges(edgeInfo, layoutInformation);

		return layoutInformation;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.impl.IGraphLayoutService#layoutGraph(java.lang.String, java.lang.String)
	 */
	@Override
	public String layoutGraph(final String nodes, final String edges) {
		final String layoutedGraph;
		final String bendPoints;

		try {
			final LayoutInformation layoutInformation = GraphLayoutServiceImpl.assembleLayoutInformation(nodes, edges);

			// Create a progress monitor
			final IKielerProgressMonitor progressMonitor = new BasicProgressMonitor();

			// Create the layout provider
			final AbstractLayoutProvider layoutProvider = new LayeredLayoutProvider();

			// Perform layout on the created graph
			layoutProvider.doLayout(layoutInformation.getGraph(), progressMonitor);

			// convert layouted graph positions to String
			layoutedGraph = GraphLayoutServiceImpl.getPositions(layoutInformation);
			bendPoints = GraphLayoutServiceImpl.getBendPoints(layoutInformation);

		} catch (final NumberFormatException e) {
			throw new GraphLayoutException("Invalid Graph Information String", e);
		}

		return layoutedGraph + "#" + bendPoints;
	}

	private static String getPositions(final LayoutInformation layoutInformation) {

		final List<List<KLabeledGraphElement>> nodesAndPorts = layoutInformation.getNodesAndPorts();
		final StringBuilder sb = new StringBuilder();
		KShapeLayout layout;
		KNode node;
		int x;
		int y; // position of node
		int width;
		int height;

		for (final List<KLabeledGraphElement> list : nodesAndPorts) {
			node = (KNode) list.get(0);
			layout = node.getData(KShapeLayout.class);
			x = Math.round(layout.getXpos() + (layout.getWidth() / 2));

			// hotFix: wrong nodeFamily position: needs to be half the height lower
			if (list.size() > 1) {
				y = Math.round(layout.getYpos() + (layout.getHeight()));
			} else {
				y = Math.round(layout.getYpos() + (layout.getHeight() / 2));
			}

			width = Math.round(layout.getWidth());
			height = Math.round(layout.getHeight());
			sb.append(x);
			sb.append(" ");
			sb.append(y);
			sb.append(" ");
			sb.append(width);
			sb.append(" ");
			sb.append(height);
			sb.append(";");
		}

		return sb.toString();
	}

	private static String getBendPoints(final LayoutInformation layoutInformation) {
		final StringBuilder sb = new StringBuilder();

		// iterate edges
		for (final KEdge edge : layoutInformation.getEdges()) {

			final KEdgeLayout layout = edge.getData(KEdgeLayout.class);
			final EList<KPoint> bendPoints = layout.getBendPoints();

			// iterate bend points
			final int bl = bendPoints.size();
			KPoint point;

			if (bl > 0) {
				point = bendPoints.get(0);
				sb.append(Math.round(point.getX()));
				sb.append(" ");
				sb.append(Math.round(point.getY()));

				for (int b = 1; b < bl; b++) {
					sb.append(" ");
					point = bendPoints.get(b);
					sb.append(Math.round(point.getX()));
					sb.append(" ");
					sb.append(Math.round(point.getY()));

				}
			}
			sb.append(";");
		}

		return sb.toString();
	}

	/**
	 * A container for the layout information which will be assembled during the layout process.
	 * 
	 * @author Nils Christian Ehmke
	 */
	private static class LayoutInformation {

		private final KNode graph;
		private final List<List<KLabeledGraphElement>> children;
		private final List<KEdge> edges;

		/**
		 * Creates a new instance of this class.
		 * 
		 * @param graph
		 *            The graph itself.
		 * @param children
		 *            The child notes and the ports.
		 * @param edges
		 *            The edges of the graph.
		 */
		public LayoutInformation(final KNode graph, final List<List<KLabeledGraphElement>> children, final List<KEdge> edges) {
			this.graph = graph;
			this.children = children;
			this.edges = edges;

			// set graph layout options
			final KShapeLayout layout = graph.getData(KShapeLayout.class);
			// layout.setProperty(LayoutOptions.LAYOUT_HIERARCHY, true);
			layout.setProperty(LayoutOptions.EDGE_ROUTING, EdgeRouting.ORTHOGONAL);

		}

		public KNode getGraph() {
			return this.graph;
		}

		public List<List<KLabeledGraphElement>> getNodesAndPorts() {
			return this.children;
		}

		public List<KEdge> getEdges() {
			return this.edges;
		}

	}
}
